from pathlib import Path


def creation_time(p: Path):
    stat = p.stat()
    if hasattr(stat, "st_birthtime"):
        return stat.st_birthtime
    return stat.st_mtime


def file_size(p: Path):
    return p.stat().st_size


def is_empty(p: Path):
    return file_size(p) == 0



from pathlib import Path
from wisort.config import loaded, overwrite_with_cli_arguments
from wisort.files import move
from wisort.files.sort import by_extension
from wisort.files.move import lib_map
from wisort.files.dedupe import dedupe
from wisort.runes import use_runes
from typing import Optional
import typer
import questionary

app = typer.Typer(
    help="""
🪄 wisort — The Magical File Sorting Companion

Sort, cleanse, and enchant your files with a touch of magic.
"""
)


def target_arg():
    return typer.Argument(
        Path.cwd(),
        help="Directory to sort (default: current directory)",
        exists=True,
        file_okay=False,
        dir_okay=True,
        resolve_path=True,
    )


def dry_run_opt():
    return typer.Option(None, "--dry-run", "-d", help="Simulate actions")


def verbose_opt():
    return typer.Option(None, "--verbose", "-v", help="Verbose output")


def force_opt():
    return typer.Option(None, "--force", "-f", help="Force moving")


def quiet_opt():
    return typer.Option(None, "--quiet", "-q", help="Suppress non-critical output")


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    if ctx.invoked_subcommand is None:
        typer.echo(ctx.get_help())
        raise typer.Exit()


@app.command(
    help="🪶 The magical apprentice that sorts your library of spellbooks and as you tell him"
)
def apprentice(
    target: Path = target_arg(),
    verbose: Optional[bool] = verbose_opt(),
    quiet: Optional[bool] = quiet_opt(),
    force: Optional[bool] = force_opt(),
):
    use_runes()
    overwrite_with_cli_arguments(quiet, verbose, force)
    if loaded.args.verbose:
        print(loaded)
    if loaded.orders.move_conflict_strategy == "mode":
        loaded.orders.move_conflict_strategy = "manual"
    if not loaded.args.quiet:
        print("""
            The apprentice stubles into your condominium sleepily.
        """)
    if len(loaded.libraries) == 0:
        if not loaded.args.quiet:
            print("""
            After a considerable amount of time searching for a library in your humble abode the apprentice realises there isn't one.
            The apprentice sighs...
            'Where does master keep sending me?', he said to himself as he left your dwelling visibly frustrated.
            """)
        return

    sorted = by_extension(target, loaded)
    l_map = lib_map(loaded.libraries)

    # future refactor maybe: replace l_map[k] by format_dest(l_map[k], path) or something to make formatting of the filename and options like flatten easier to implement
    dest_map = {
        path.absolute(): l_map[k]
        for k, paths in sorted.items()
        if k in l_map
        for path in paths
    }

    if len(dest_map) == 0:
        if not loaded.args.quiet:
            print("""
            'Damn this place is very tidy', he thought.
                  """)
            return

    if not loaded.args.quiet:
        print(f"""
            Your {len(loaded.libraries.keys())} libraries excite the apprentice as he gazes at the rows of neatly stored books in all the bookshelves
        """)
        print("""
            After a brief moment of reviewing your books he snorts.
            'Oh man the people nowadays really don't have structure in their libraries. I can see why my master sent me'
            """)
        print(f"""
            Ensuing a quick inspection of your spellbooks he realised there where only {len(dest_map)} books that needed organizing.
        """)

    if loaded.args.verbose:
        for path, dest in dest_map.items():
            print(f"""          {path.relative_to(target)} → {dest}""")

    if not loaded.args.quiet:
        print("""
            He remembered a nice spell his master thaught him the previous day to levitate and move objects.
            After an astoundingly high number of attempts he finally managed to move all books to their destinations using the spell. 
            Unsatisfied he sighed: 'If only I had done this normally without spells I would long be home by now.'.
            """)

    if not loaded.args.quiet:
        print("""
            The apprentice comes up to you and asks:
            'Can I organize your libraries?'
            """)

    run = questionary.confirm("Start file organization?").ask()
    if not run:
        return

    move(dest_map, target, loaded)

    if loaded.orders.dedupe:
        pass

    return


@app.command(
    help="🔮 Keeps your spellbooks and library organized for you as he was thaught."
)
def magician(
    target: Path = target_arg(),
    verbose: bool = verbose_opt(),
    quiet: bool = quiet_opt(),
):
    overwrite_with_cli_arguments(quiet, verbose, None)
    if loaded.orders.move_conflict_strategy == "mode":
        loaded.orders.move_conflict_strategy = "rename"
    if not loaded.args.quiet:
        print("POOOOOOOOF! A magician appears out of thin air.")

    print("The magician is still an apprentice. He hasn't finished his studies yet")


@app.command(help="🧹 Full spellbook: sort, dedupe, and format code")
def witch(
    target: Path = target_arg(),
    verbose: bool = verbose_opt(),
    quiet: bool = quiet_opt(),
):
    overwrite_with_cli_arguments(quiet, verbose, None)
    use_runes()
    if loaded.args.verbose:
        print(loaded)
    dedupe(target, loaded)



from wisort.config import Library


def lib_map(libraries: dict[str, Library]) -> dict[str, Library]:
    result: dict[str, str] = {}
    for lib in libraries.values():
        for ft in lib.filetypes:
            result[ft] = lib

    return result



from pathlib import Path
from wisort.config import Config, Library
from wisort.files.conflicts import move_conflict_resolution
import questionary


def move(map: dict[Path, Library], target: Path, cfg: Config):
    for src, lib in map.items():
        dest = Path(lib.destination).expanduser().absolute()

        if not dest.is_dir():
            raise Exception(
                f"Destination path {
                    dest.absolute()
                } is not a directory or does not exist"
            )

        # do some smart automations
        # dont move if in ignored
        # handle symlinks
        # unzip on move? -> smartly if zip has multiple elements zip to folder, otherwise zip to element directly

        # this doesnt support preserve file structure as of now

        parent: Path = dest
        name = src.name

        if (
            lib.flatten is not None and not lib.flatten
        ) or cfg.orders.move_strategy == "preserveFolders":
            parent = dest / src.relative_to(target).parent
            parent.mkdir(exist_ok=True, parents=True)

        df = parent / name

        if df.exists():
            # conflict resolution
            (df, m) = move_conflict_resolution(name, parent, target, dest, cfg)
            if not m:
                if cfg.args.verbose:
                    print(f"skipped {df.relative_to(dest)}")
                continue
        src.move(df)



from wisort.config import Config
from pathlib import Path
import questionary


def move_conflict_resolution(
    name: str, parent: Path, target: Path, dest: Path, cfg: Config
) -> (Path, bool):
    def dfile() -> Path:
        return parent / name

    match cfg.orders.move_conflict_strategy:
        case "manual":
            mode = questionary.select(
                f"{dfile().relative_to(dest)} already exists in {
                    dest
                }. What do you want to do?",
                choices=["remove", "skip", "rename"],
            ).ask()
            ncfg = cfg
            ncfg.orders.move_conflict_strategy = mode
            return move_conflict_resolution(name, parent, target, dest, cfg)
        case "remove":
            return (dfile(), True)
        case "skip":
            return (dfile(), False)
        case "rename":
            name = questionary.text(
                f"A file named {dfile().relative_to(dest)} already exists. Rename to:",
                validate=lambda n: not dfile().exists(),
            ).ask()
            return (dfile(), True)
        case _:
            raise Exception(
                f"`{cfg.orders.move_conflict_strategy}` is only a placeholder move conflict resolution strategy"
            )



from wisort.config import Config, loaded


def use_runes():
    for name, rune in loaded.runes.items():
        if not isinstance(rune, str):
            continue
        if rune.startswith("@"):
            if rune[1:] not in loaded.runes:
                print(f"`rune {rune[1:]}` not found in your config")
                continue
            loaded.runes[name] = loaded.runes[rune[1:]]
            # TODO: add type checking
    for name, lib in loaded.libraries.items():
        for attr, value in vars(lib).items():
            if not isinstance(value, str):
                continue
            if value.startswith("@"):
                if value[1:] not in loaded.runes:
                    print(f"`rune {value[1:]}` not found in your config")
                    continue
                setattr(loaded.libraries[name], attr, loaded.runes[value[1:]])
                # TODO: add type checking
    for attr, value in vars(loaded.orders).items():
        if not isinstance(value, str):
            continue
        if value.startswith("@"):
            if value[1:] not in loaded.runes:
                print(f"`rune {value[1:]}` not found in your config")
                continue
            setattr(loaded.libraries[name], attr, loaded.runes[value[1:]])
            # TODO: add type checking



from wisort.config import Config
from wisort.files.utils import creation_time, is_empty
from pathlib import Path
from blake3 import blake3


def find_dupes(target: Path, cfg: Config) -> dict[str, list[Path]]:
    dupes: dict[str, list[Path]] = {}
    for file in target.iterdir():
        if file.is_symlink():
            continue
        if file.is_dir() and cfg.orders.recurse:
            ds = find_dupes(file, cfg)
            for key, d in ds.items():
                dupes.setdefault(key, [d]).extend(d)
            continue

        if is_empty(file) and cfg.orders.delete_empty_files:
            file.unlink()
            continue
        file_bytes = file.read_bytes()
        digest = blake3(file_bytes).hexdigest()

        dupes.setdefault(digest, [file]).append(file)

    return dupes


def dedupe(target: Path, cfg: Config):
    dupes = find_dupes(target, cfg)
    n = 0
    for digest, duplicates in dupes.items():
        if len(duplicates) == 1:
            continue
        n += len(duplicates) - 1
        sorted_dupes = sorted(duplicates, key=lambda p: creation_time(p))
        original = sorted_dupes[0].resolve()
        for dupe in sorted_dupes:
            if original == dupe:
                continue
            match cfg.orders.dedupe_strategy:
                case "portal":
                    dupe.unlink()
                    dupe.symlink_to(sorted_dupes[0])
                    if cfg.args.verbose:
                        print(
                            f"symlinked {dupe.relative_to(target)} to {sorted_dupes[0]}"
                        )

                case "remove":
                    if cfg.args.verbose:
                        print(f"deleted {dupe.relative_to(target)}")
                    dupe.unlink()
    if cfg.args.verbose:
        print(f"{n} duplicates where taken care of")



from pathlib import Path, PurePath
from pathspec import PathSpec
from wisort.config import Config


def by_extension(
    directory: Path,
    cfg: Config,
    spec: PathSpec = None,
) -> dict[str, list[Path]]:
    if spec is None:
        spec = PathSpec.from_lines("gitwildmatch", [])
    paths: dict[str, list[Path]] = {}
    if not directory.is_dir():
        return []

    # copy specs -> otherwise child .gitignores mutate parent gitignores because of recursion
    local_spec = PathSpec.from_lines("gitwildmatch", list(spec.patterns))

    for entry in directory.iterdir():
        if (
            entry.name == ".gitignore"
            and cfg.orders.honor_gitignore
            and not entry.is_dir()
        ):
            if cfg.args.verbose:
                print(f"reading {entry.absolute()}")
            patterns = entry.read_text().splitlines()
            local_spec += PathSpec.from_lines("gitwildmatch", patterns)
            continue

        # Ignore dotfiles
        if entry.name[0] == "." and cfg.orders.ignore_dotfiles:
            continue

        # recursively go through subdirs
        if entry.is_dir() and cfg.orders.recurse:
            ps = by_extension(entry, cfg, local_spec)
            for key, p in ps.items():
                paths.setdefault(key, []).extend(p)
            continue

        in_gitignore = local_spec.match_file(str(entry.relative_to(directory)))
        if len(entry.suffix) != 0 and not (in_gitignore and cfg.orders.honor_gitignore):
            paths.setdefault(entry.suffix[1:], []).append(entry)

    return paths



from pathlib import Path
from pydantic import BaseModel, Field
from typing import Optional, Literal
import json
from os import environ


class Library(BaseModel):
    destination: str
    filetypes: list[str] | str
    flatten: Optional[bool] | str = Field(default=None)


class Orders(BaseModel):
    unzip: bool = Field(default=True)
    dedupe: bool = Field(default=False)

    # portal: symlinks
    # remove: removes
    # auto: symlinks if duplicate looks intentional: completely other name, etc.
    dedupe_strategy: Literal["portal", "remove", "auto"] = Field(default="portal")

    recurse: bool = Field(default=True)
    move_strategy: Literal["preserveFolders", "flatten"] = Field(default="flatten")
    move_conflict_strategy: Literal["manual", "remove", "skip", "rename", "mode"] = (
        Field(default="mode")
    )
    honor_gitignore: bool = Field(default=True)
    ignore_dotfiles: bool = Field(default=True)
    delete_empty_files: bool = Field(default=True)


class Arguments(BaseModel):
    quiet: bool = Field(default=False)
    verbose: bool = Field(default=False)
    force: bool = Field(default=False)


class Config(BaseModel):
    # spell = spell (dict[str])
    # spell = list of str -> e.g. list of filetypes
    # ...
    runes: dict[str, str | list[str] | bool | int]
    libraries: dict[str, Library] | list[Library]
    orders: Orders
    args: Arguments


def load(
    path: str = "./config.json",
) -> Config:
    with open(path, "r") as f:
        data = json.load(f)
    return Config(**data)


# TODO: make this into decorator that wraps name?
def overwrite_with_cli_arguments(
    quiet: Optional[bool], verbose: Optional[bool], force: Optional[bool]
):
    if quiet is not None:
        loaded.args.quiet = quiet
    if verbose is not None:
        loaded.args.verbose = verbose
    if force is not None:
        loaded.orders.move_conflict_strategy = "remove"
        loaded.args.force = force


config_home = (
    Path(environ["XDG_CONFIG_HOME"])
    if "XDG_CONFIG_HOME" in environ
    else Path(environ["HOME"] / ".config")
)
loaded = load(Path(environ["XDG_CONFIG_HOME"]) / "wisort/config.json")
